/*********************************************************************/
/* file: color.c - functions related to the color command            */
/*                             HUT TINTIN                            */
/*                        coded by Ossi Tulonen                      */
/*********************************************************************/
/*
 * Copyright (c) 1992 1993 1994 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
 
#include <stdio.h>
#include <string.h>
#ifndef NSTDLIBS
#include <stdlib.h>
#endif

#include "structs.h"
#include "color.h"
#include "llist.h"
#include "parse.h"
#include "screen.h"
#include "utility.h"
#include "utils.h"

extern struct master_table_data master_table;

extern const uchar *RESETCOLOR;

const unsigned char *colors[] = {
  (const uchar *)"reset",          (const uchar *)"bold",         /*  0 -  1 */
  (const uchar *)"faint",          (const uchar *)"italicised",
  (const uchar *)"underlined",     (const uchar *)"blink",
  (const uchar *)"",               (const uchar *)"negative",
  (const uchar *)"",               (const uchar *)"",             /*  8 -  9 */
  (const uchar *)"",               (const uchar *)"",
  (const uchar *)"",               (const uchar *)"",
  (const uchar *)"",               (const uchar *)"",
  (const uchar *)"",               (const uchar *)"",
  (const uchar *)"",               (const uchar *)"",             /* 18 - 19 */
  (const uchar *)"",               (const uchar *)"",
  (const uchar *)"bold_off",       (const uchar *)"italicised_off",
  (const uchar *)"underlined_off", (const uchar *)"blink_off",
  (const uchar *)"",               (const uchar *)"negative_off",
  (const uchar *)"",               (const uchar *)"",             /* 28 - 29 */
  (const uchar *)"black",          (const uchar *)"red",
  (const uchar *)"green",          (const uchar *)"yellow",
  (const uchar *)"blue",           (const uchar *)"violet",
  (const uchar *)"cyan",           (const uchar *)"white",
  (const uchar *)"",               (const uchar *)"",             /* 38 - 39 */
  (const uchar *)"BLACK",          (const uchar *)"RED",
  (const uchar *)"GREEN",          (const uchar *)"YELLOW",
  (const uchar *)"BLUE",           (const uchar *)"VIOLET",
  (const uchar *)"CYAN",           (const uchar *)"WHITE",
  (const uchar *)"",               (const uchar *)""              /* 45 - 49 */
};

/* ************************************************************
   the #color command
************************************************************ */

struct session *comm_color(const unsigned char *arg, session *ses)
{
  unsigned char left[BUFFER_SIZE], right[BUFFER_SIZE], buf[BUFFER_SIZE];
  int i, j, pos = 0, found;
  struct listnode **mycolors, **myvars, *ln;

  if(strcmp((const char *)arg, "help") == 0) {
    ses->print((const uchar *)"#COLORS:\n\r");
    for(i = 0, j = 0; i < 48; i++) {
      if(*colors[i]) {
        if(!j)
          pos = str_cat(buf, 0, sizeof(buf), (const uchar *)"        ");

        sprintf((char *)left, "  \033[%dm%10s%s", i, colors[i], RESETCOLOR);
        pos = str_cat(buf, pos, sizeof(buf), left);
        if(++j == 5) {
	  pos = str_cat(buf, pos, sizeof(buf), (const uchar *)"\n\r");
          ses->print(buf);
          j = 0;
        }
      }
    }
    if(j) {
      pos = str_cat(buf, pos, sizeof(buf), (const uchar *)"\n\r");
      ses->print(buf);
    }
    return(ses);
  }

  myvars = &ses->variables;
  mycolors = &ses->colors;

  arg = get_arg_stop_spaces(arg, left);
  arg = get_arg_with_spaces(arg, right);

  if(!*left) {
    ses->print((const uchar *)"#THESE COLOR VARIABLES HAS BEEN DEFINED:\n\r");
    show_list(*mycolors, ses);
 } else if(!*right) {
    for(found = FALSE,ln = searchnode_list2(*mycolors, left); ln;
                      ln = searchnode_list2(ln->next, left)) {
      found = TRUE;
      ln->show(ses);
    }
    if(!found)
      ses->print((const uchar *)"#THAT COLOR VARIABLE ISN'T DEFINED.\n\r");
  } else {
    text_to_color(right, buf, ses);

    if(*buf) {
      if((ln = searchnode_list(*myvars, left)) != NULL)
        delete ln;
      
      if((ln = searchnode_list(*mycolors, left)) != NULL)
        delete ln;
      ln = new listnode(mycolors, left, buf, TYPE_COLOR, 5);
      if(IS_SET(ses->message_flags, MESSAGE_VAR))
	ses->print((const uchar *)"#OK. COLOR VARIABLE DEFINED.\n\r");
    }
  }
  return(ses);
}

void color_to_text(unsigned char *color, unsigned char *text, int flag)
{
  unsigned char buf[10], *tmp, buf2[10];
  int pos, multi, colornum;

  color += 2;  /* remove esc and [ */

  multi = FALSE;
  tmp = buf;
  pos = 0;
  while(*color) {
    if(*color == ';' || *color == 'm') {
      *tmp = '\0';
      colornum = atoi((const char *)buf);
      if(multi)
        pos = str_cat(text, pos, BUFFER_SIZE, (const uchar *)" ");
      else
        multi = TRUE;

      if(flag) {
        sprintf((char *)buf2, "\033[%dm", colornum);
        pos = str_cat(text, pos, BUFFER_SIZE, buf2);
      }
      pos = str_cat(text, pos, BUFFER_SIZE, colors[colornum]);
      if(flag)
        pos = str_cat(text, pos, BUFFER_SIZE, RESETCOLOR);

      tmp = buf;
      color++;
    } else {
      *tmp++ = *color++;
    }
  }
}

void text_to_color(const unsigned char *text, unsigned char *color,
		   struct session *ses)
{
  unsigned char buf[BUFFER_SIZE], *tmp, buf2[BUFFER_SIZE];
  int pos, multi, i;

  pos = str_cat(color, 0, BUFFER_SIZE, (const uchar *)"\033[");

  multi = FALSE;
  text = space_out(text);
  while(*text) {
    tmp = buf;
    while(*text && *text != ' ')
      *tmp++ = *text++;
    *tmp = '\0';

    for(i = 0; i < 48; i++)
      if(strcmp((const char *)colors[i], (const char *)buf) == 0)
        break;

    if(i == 48) {
      sprintf((char *)buf2, "#NO SUCH COLOR (%s).\n\r", buf);
      ses->print(buf2);
    } else {
      if(multi)
        pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)";");
      else
        multi = TRUE;

      sprintf((char *)buf2, "%d", i);
      pos = str_cat(color, pos, BUFFER_SIZE, buf2);
    }
    text = space_out(text);
  }

  pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"m");
  if(!multi)
    *color = '\0';
}

void color_to_bits(const unsigned char *color, unsigned char *bits)
{
  unsigned char buf[10], *tmp;
  int colornum;

  color += 2;  /* remove esc and [ */
  
  tmp = buf;
  while(*color) {
    if(*color == ';' || *color == 'm') {
      *tmp = '\0';
      colornum = atoi((const char *)buf);
      
      if(colornum == 0) {
	bits[0] = 0x01;
	bits[1] = 0;
	bits[2] = 0;
      } else {
	bits[0] &= ~0x01;
	switch(colornum) {
	case 1:
	  bits[0] |= 0x02;
	  bits[0] &= ~0x04;
	  break;
	case 2:
	  bits[0] |= 0x04;
	  bits[0] &= ~0x02;
	  break;
	case 3:
	  bits[0] |= 0x08;
	  break;
	case 4:
	  bits[0] |= 0x10;
	  break;
	case 5:
	  bits[0] |= 0x20;
	  break;
	case 7:
	  bits[0] |= 0x80;
	  break;
	case 22:
	  bits[0] &= ~0x02;
	  break;
	case 23:
	  bits[0] &= ~0x08;
	case 24:
	  bits[0] &= ~0x10;
	case 25:
	  bits[0] &= ~0x20;
	case 27:
	  bits[0] &= ~0x80;
	  break;
	case 30:
	case 31:
	case 32:
	case 33:
	case 34:
	case 35:
	case 36:
	case 37:
	  bits[1] = colornum;
	  break;
	case 40:
	case 41:
	case 42:
	case 43:
	case 44:
	case 45:
	case 46:
	case 47:
	  bits[2] = colornum;
	  break;
	default:
	  break;
	}
      }
      tmp = buf;
      color++;
    } else {
      *tmp++ = *color++;
    }
  }
}

unsigned char *bits_to_color(const unsigned char *obits,
			     const unsigned char *nbits,
			     unsigned char *color)
{
  unsigned char tmp[20], bits[4];
  int pos = 0;
  
  *color = '\0';
  if(obits[0] == nbits[0] &&
     obits[1] == nbits[1] &&
     obits[2] == nbits[2])
    return(color);
  
  pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"\033[");
  if(nbits[0] == 0x01) {
    pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"0m");
  } else {
    if((obits[1] != nbits[1] && nbits[1] == 0) ||
       (obits[2] != nbits[2] && nbits[2] == 0) ||
       (master_table.no_neg_off && !(nbits[0]&0x80) && (obits[0]&0x80)) ||
       (!(nbits[0]&0x04) && (obits[0]&0x04))) {
      pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"0;");
      bits[0] = 0;
      bits[1] = 0;
      bits[2] = 0;
      bits[3] = 0;
    } else {
      bits[0] = obits[0];
      bits[1] = obits[1];
      bits[2] = obits[2];
      bits[3] = obits[3];
    }
    
    if(bits[0] != nbits[0]) {
      // bold
      if((nbits[0]&0x02) && !(bits[0]&0x02))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"1;");
      if(!(nbits[0]&0x02) && (bits[0]&0x02))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"22;");
      
      // faint
      if((nbits[0]&0x04) && !(bits[0]&0x04))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"2;");
      
      // italiced
      if((nbits[0]&0x08) && !(bits[0]&0x08))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"3;");
      if(!(nbits[0]&0x08) && (bits[0]&0x08))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"23;");
      
      // underlined
      if((nbits[0]&0x10) && !(bits[0]&0x10))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"4;");
      if(!(nbits[0]&0x10) && (bits[0]&0x10))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"24;");
      
      // blink
      if((nbits[0]&0x20) && !(bits[0]&0x20))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"5;");
      if(!(nbits[0]&0x20) && (bits[0]&0x20))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"25;");
      
      // negative
      if((nbits[0]&0x80) && !(bits[0]&0x80))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"7;");
      if(!(nbits[0]&0x80) && (bits[0]&0x80))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"27;");
    }
    // text color
    if(bits[1] != nbits[1]) {
      sprintf((char *)tmp, "%d", (int)nbits[1]);
      pos = str_cat(color, pos, BUFFER_SIZE, tmp);
      pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)";");
    }
    // background color
    if(bits[2] != nbits[2]) {
      sprintf((char *)tmp, "%d", (int)nbits[2]);
      pos = str_cat(color, pos, BUFFER_SIZE, tmp);
      pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"m");
    } else {
      pos = str_cat(color, pos - 1, BUFFER_SIZE, (const uchar *)"m");
    }
  }
  return(color);
}
